@@ -4,6 +4,10 @@ use std::ffi::{ CStr };
 use std::ptr;
 use std::ffi;
 use std::collections::HashMap;
+use std::pin::Pin;
+
+use urid::*;
+use lv2_urid::*;
 
 // find plugins in /usr/lib/lv2
 // URI list with lv2ls
@@ -13,7 +17,7 @@ use std::collections::HashMap;
 // https://discord.com/channels/273534239310479360/592856094527848449/837709621111947285
 
 const LILV_URI_CONNECTION_OPTIONAL: &[u8; 48usize] = b"http://lv2plug.in/ns/lv2core#connectionOptional\0";
-pub const LV2_URID__map: &[u8; 34usize] = b"http://lv2plug.in/ns/ext/urid#map\0";
+pub const LV2_URID_MAP: &[u8; 34usize] = b"http://lv2plug.in/ns/ext/urid#map\0";
 
 pub struct Lv2Host{
     world: *mut LilvWorld,
@@ -23,14 +27,46 @@ pub struct Lv2Host{
     uri_home: Vec<String>,
     plugin_names: HashMap<String, usize>,
     dead_list: Vec<usize>,
+    #[allow(dead_code)]
     buffer_len: usize,
+    sr: f64,
     in_buf: Vec<f32>,
     out_buf: Vec<f32>,
     atom_buf: [u8; 1024],
+    atom_urid_bytes: [u8; 4],
+    midi_urid_bytes: [u8; 4],
+    #[allow(dead_code)]
+    host_map: Pin<Box<HostMap<HashURIDMapper>>>,
+    pub map_interface: lv2_sys::LV2_URID_Map,
+}
+
+#[derive(Clone, Copy, PartialEq, Eq, Debug)]
+pub enum AddPluginError{
+    CapacityReached,
+    MoreThanTwoInOrOutAudioPorts(usize, usize),
+    MoreThanOneAtomPort(usize),
+    WorldIsNull,
+    PluginIsNull,
+    PortNeitherInputOrOutput,
+    PortNeitherControlOrAudioOrOptional,
+}
+
+#[derive(Clone, Copy, PartialEq, Eq, Debug)]
+pub enum ApplyError{
+    PluginIndexOutOfBound,
+    LeftRightInputLenUnequal,
+    MoreFramesThanBufferLen,
 }
 
 impl Lv2Host{
-    pub fn new(plugin_cap: usize, buffer_len: usize) -> Self{
+    pub fn new(plugin_cap: usize, buffer_len: usize, sample_rate: usize) -> Self{
+        // setup feature map
+        let mut host_map: Pin<Box<HostMap<HashURIDMapper>>> = Box::pin(HashURIDMapper::new().into());
+        let map_interface = host_map.as_mut().make_map_interface();
+        let map = LV2Map::new(&map_interface);
+        let midi_urid_bytes = map.map_str("http://lv2plug.in/ns/ext/midi#MidiEvent").unwrap().get().to_le_bytes();
+        let atom_urid_bytes = map.map_str("http://lv2plug.in/ns/ext/atom#Sequence").unwrap().get().to_le_bytes();
+
         let (world, lilv_plugins) = unsafe{
             let world = lilv_world_new();
             lilv_world_load_all(world);
@@ -46,61 +82,34 @@ impl Lv2Host{
             plugin_names: HashMap::new(),
             dead_list: Vec::new(),
             buffer_len,
+            sr: sample_rate as f64,
             in_buf: vec![0.0; buffer_len * 2], // also don't let it resize
             out_buf: vec![0.0; buffer_len * 2], // also don't let it resize
             atom_buf: [0; 1024],
+            atom_urid_bytes,
+            midi_urid_bytes,
+            host_map,
+            map_interface,
         }
     }
 
     pub fn get_index(&self, name: &str) -> Option<usize>{
-        if let Some(index) = self.plugin_names.get(name){
-            Some(*index)
-        } else {
-            None
-        }
-    }
-
-    pub fn test_midi_atom(&mut self, typebytes: [u8; 4], seqbytes: [u8; 4]){
-        // header seq urid
-        self.atom_buf[0] = seqbytes[0];
-        self.atom_buf[1] = seqbytes[1];
-        self.atom_buf[2] = seqbytes[2];
-        self.atom_buf[3] = seqbytes[3];
-        // size
-        self.atom_buf[4] = 0;
-        self.atom_buf[5] = 0;
-        self.atom_buf[6] = 0;
-        self.atom_buf[7] = 24;
-        // timestamp
-        self.atom_buf[8] = 0;
-        self.atom_buf[9] = 0;
-        self.atom_buf[10] = 0;
-        self.atom_buf[11] = 0;
-        self.atom_buf[12] = 0;
-        self.atom_buf[13] = 0;
-        self.atom_buf[14] = 0;
-        self.atom_buf[15] = 0;
-        // type
-        self.atom_buf[16] = typebytes[0];
-        self.atom_buf[17] = typebytes[1];
-        self.atom_buf[18] = typebytes[2];
-        self.atom_buf[19] = typebytes[3];
-        // size
-        self.atom_buf[20] = 3;
-        self.atom_buf[21] = 0;
-        self.atom_buf[22] = 0;
-        self.atom_buf[23] = 0;
-        // midi on
-        self.atom_buf[24] = 0x90;
-        self.atom_buf[25] = 50;
-        self.atom_buf[26] = 100;
+        self.plugin_names.get(name).copied()
     }
 
     #[allow(clippy::not_unsafe_ptr_arg_deref)]
-    pub fn add_plugin(&mut self, uri: &str, name: String, features_ptr: *const *const lv2_raw::core::LV2Feature) -> Result<(), String>{
+    pub fn add_plugin(&mut self, uri: &str, name: String) -> Result<usize, AddPluginError>{
+        let feature_vec = vec![lv2_raw::core::LV2Feature {
+            uri: LV2_URID_MAP.as_ptr() as *const i8,
+            data: &mut self.map_interface as *mut lv2_sys::LV2_URID_Map as *mut std::ffi::c_void,
+        }];
+        let mapfp = feature_vec.as_ptr() as *const lv2_raw::core::LV2Feature;
+        let features = vec![mapfp, std::ptr::null::<lv2_raw::core::LV2Feature>()];
+
+        let features_ptr = features.as_ptr() as *const *const lv2_raw::core::LV2Feature;
         let replace_index = self.dead_list.pop();
         if self.plugins.len() == self.plugin_cap && replace_index == None{
-            return Err("Lv2mh: can't add plugin: all plugin capacity used.".to_owned());
+            return Err(AddPluginError::CapacityReached);
         }
         let mut uri_src = uri.to_owned();
         uri_src.push('\0'); // make sure it's null terminated
@@ -112,20 +121,23 @@ impl Lv2Host{
             plugin
         };
 
-        let (mut ports, port_names, n_audio_in, n_audio_out, n_atom_in) = unsafe{ create_ports(self.world, plugin) };
+        let (mut ports, port_names, n_audio_in, n_audio_out, n_atom_in) = unsafe {
+            create_ports(self.world, plugin)?
+        };
+
         if n_audio_in > 2 || n_audio_out > 2 {
-            return Err(format!("Lv2hm: can only have at max 2 audio input ports and 2 audio output ports. \n\t Audio input ports: {}, Audio output ports: {}", n_audio_in, n_audio_out));
+            return Err(AddPluginError::MoreThanTwoInOrOutAudioPorts(n_audio_in, n_audio_out));
         }
 
         if n_atom_in > 1 {
-            return Err(format!("Lv2hm: plugin has more than one atom input port: {} atom input ports found", n_atom_in));
+            return Err(AddPluginError::MoreThanOneAtomPort(n_atom_in));
         }
 
         let instance = unsafe{
-            let map_feature_uri = lilv_new_uri(self.world, LV2_URID__map.as_ptr() as *const i8);
-            let map_feature_bool = lilv_plugin_has_feature(plugin, map_feature_uri);
-            println!("supports map: {}", map_feature_bool); // prints true, supports map
-            let instance = lilv_plugin_instantiate(plugin, 44100.0, features_ptr);
+            //let map_feature_uri = lilv_new_uri(self.world, LV2_URID_MAP.as_ptr() as *const i8);
+            //let map_feature_bool = lilv_plugin_has_feature(plugin, map_feature_uri);
+            //println!("supports map: {}", map_feature_bool); // prints true, supports map
+            let instance = lilv_plugin_instantiate(plugin, self.sr, features_ptr);
             let (mut i, mut o) = (0, 0);
             for p in &mut ports{
                 match p.ptype{
@@ -134,10 +146,10 @@ impl Lv2Host{
                     },
                     PortType::Audio => {
                         if p.is_input {
-                            lilv_instance_connect_port(instance, p.index, self.in_buf.as_ptr().offset(i) as *mut ffi::c_void);
+                            lilv_instance_connect_port(instance, p.index, self.in_buf.as_ptr().offset(i*self.buffer_len as isize) as *mut ffi::c_void);
                             i += 1;
                         } else {
-                            lilv_instance_connect_port(instance, p.index, self.out_buf.as_ptr().offset(o) as *mut ffi::c_void);
+                            lilv_instance_connect_port(instance, p.index, self.out_buf.as_ptr().offset(o*self.buffer_len as isize) as *mut ffi::c_void);
                             o += 1;
                         }
                     },
@@ -155,7 +167,6 @@ impl Lv2Host{
         };
 
         let p = Plugin{
-            lilv_plugin: plugin,
             instance,
             port_names,
             ports,
@@ -164,13 +175,12 @@ impl Lv2Host{
         if let Some(i) = replace_index{
             self.plugins[i] = p;
             self.plugin_names.insert(name, i);
-            println!("dstn: {}", i);
         } else {
             self.plugins.push(p);
             self.plugin_names.insert(name, self.plugins.len() - 1);
         }
 
-        Ok(())
+        Ok(self.plugins.len() - 1)
     }
 
     pub fn remove_plugin(&mut self, name: &str) -> bool{
@@ -241,42 +251,75 @@ impl Lv2Host{
         }
     }
 
-    pub fn apply_plugin(&mut self, index: usize, input_frame: (f32, f32)) -> (f32, f32){
+    pub fn apply(&mut self, index: usize, input: [u8; 3], input_frame: (f32, f32)) -> (f32, f32) {
         if index >= self.plugins.len() { return (0.0, 0.0); }
-        let plugin = &mut self.plugins[index];
         self.in_buf[0] = input_frame.0;
         self.in_buf[1] = input_frame.1;
+        let plugin = &mut self.plugins[index];
+        midi_into_atom_buffer(self.midi_urid_bytes, self.atom_urid_bytes, vec![(0, input)], &mut self.atom_buf);
         unsafe {
             lilv_instance_run(plugin.instance, 1);
         }
         (self.out_buf[0], self.out_buf[1])
     }
 
-    // not tested yet, need to get more of the daw up
-    pub fn apply_plugin_n_frames(&mut self, index: usize, input: &[f32]) -> Option<&[f32]>{
-        let frames = input.len() / 2;
-        if frames > self.buffer_len { return None; }
-        if index >= self.plugins.len() { return None; }
-        for (i, v) in input.iter().enumerate(){
-            self.in_buf[i] = *v;
+    pub fn apply_multi(&mut self, index: usize, input: Vec<(u64, [u8; 3])>, input_frame: [&[f32]; 2]) -> Result<[&[f32]; 2], ApplyError>{
+        midi_into_atom_buffer(self.midi_urid_bytes, self.atom_urid_bytes, input, &mut self.atom_buf);
+        if index >= self.plugins.len() { return Err(ApplyError::PluginIndexOutOfBound); }
+        let frames = input_frame[0].len();
+        if frames != input_frame[1].len(){
+            return Err(ApplyError::LeftRightInputLenUnequal);
+        }
+        if frames > self.buffer_len{
+            return Err(ApplyError::MoreFramesThanBufferLen);
+        }
+        for (i, (l, r)) in input_frame[0].iter().zip(input_frame[1].iter()).enumerate(){
+            self.in_buf[i] = *l;
+            self.in_buf[i + self.buffer_len] = *r;
         }
         let plugin = &mut self.plugins[index];
         unsafe{
             lilv_instance_run(plugin.instance, frames as u32);
         }
-        Some(&self.out_buf)
-    }
-
-    pub fn apply_instrument(&mut self, index: usize) -> (f32, f32){
-        if index >= self.plugins.len() { return (0.0, 0.0); }
-        let plugin = &mut self.plugins[index];
-        unsafe {
-            lilv_instance_run(plugin.instance, 1);
-        }
-        (self.out_buf[0], self.out_buf[1])
+        Ok([&self.out_buf[..frames], &self.out_buf[self.buffer_len..][..frames]])
     }
 }
 
+fn midi_into_atom_buffer(typebytes: [u8; 4], seqbytes: [u8; 4], midibytes: Vec<(u64, [u8; 3])>, atom_buf: &mut [u8; 1024]) {
+    // size gets written at the end
+    // SELF.ATOM_BUF[0..4] = [8,0,0,0];
+    let mut pos = 4; // current offset
+    // type
+    copy_bytes(atom_buf, &seqbytes, &mut pos); // type: sequence
+    copy_bytes(atom_buf, &[0,0,0,0,0,0,0,0,], &mut pos); // frame
+
+    for (ea_time, ea_midi) in midibytes {
+        // timestamp
+        copy_bytes(atom_buf, &ea_time.to_le_bytes(), &mut pos); // subframe
+
+        copy_bytes(atom_buf, &3_u32.to_le_bytes(), &mut pos); // size
+        copy_bytes(atom_buf, &typebytes, &mut pos); // type: midi
+        copy_bytes(atom_buf, &ea_midi, &mut pos);
+        let extra = 8 - (pos % 8);
+        if extra != 8 {
+            for idx in 0..extra {
+                atom_buf[pos..][idx] = 0;
+            }
+            pos += extra;
+        }
+    }
+    pos -= 8; // length shouldn't include size and type of top-level atom
+    copy_bytes(atom_buf, &(pos as u32).to_le_bytes(), &mut 0); // size
+}
+
+fn copy_bytes(to: &mut [u8], from: &[u8], at: &mut usize) {
+    for (to, from) in to[*at..].iter_mut().zip(from.iter()) {
+        *to = *from;
+    }
+    *at += from.len();
+}
+
+
 impl Drop for Lv2Host{
     fn drop(&mut self){
         unsafe{
@@ -293,7 +336,6 @@ impl Drop for Lv2Host{
 }
 
 struct Plugin{
-    lilv_plugin: *const LilvPlugin,
     instance: *mut LilvInstance,
     port_names: HashMap<String, usize>,
     ports: Vec<Port>,
@@ -323,10 +365,12 @@ struct Port{
     name: String,
 }
 
+type CreatePortsRes = (Vec<Port>, HashMap<String, usize>, usize, usize, usize);
+
 // ([ports], [(port_name, index)], n_audio_in, n_audio_out, n_atom_in)
-unsafe fn create_ports(world: *mut LilvWorld, plugin: *const LilvPluginImpl) -> (Vec<Port>, HashMap<String, usize>, usize, usize, usize){
-    if world.is_null() { panic!("Lv2hm: create_ports: world is null."); }
-    if plugin.is_null() { panic!("Lv2hm: create_ports: plugin is null."); }
+unsafe fn create_ports(world: *mut LilvWorld, plugin: *const LilvPluginImpl) -> Result<CreatePortsRes, AddPluginError>{
+    if world.is_null() { return Err(AddPluginError::WorldIsNull); }
+    if plugin.is_null() { return Err(AddPluginError::PluginIsNull); }
 
     let mut ports = Vec::new();
     let mut names = HashMap::new();
@@ -365,7 +409,7 @@ unsafe fn create_ports(world: *mut LilvWorld, plugin: *const LilvPluginImpl) ->
         let optional = lilv_port_has_property(plugin, lport, lv2_connection_optional);
 
         let is_input = if lilv_port_is_a(plugin, lport, lv2_input_port) { true }
-        else if !lilv_port_is_a(plugin, lport, lv2_output_port) && !optional { panic!("Lv2hm: Port is neither input nor output."); }
+        else if !lilv_port_is_a(plugin, lport, lv2_output_port) && !optional { return Err(AddPluginError::PortNeitherInputOrOutput); }
         else { false };
 
         let ptype = if lilv_port_is_a(plugin, lport, lv2_control_port) { PortType::Control }
@@ -381,7 +425,7 @@ unsafe fn create_ports(world: *mut LilvWorld, plugin: *const LilvPluginImpl) ->
             n_atom_in += 1;
             PortType::Atom
         }
-        else if !optional { panic!("Lv2hm: port is neither a control, audio or optional port."); }
+        else if !optional { return Err(AddPluginError::PortNeitherControlOrAudioOrOptional); }
         else { PortType::Other };
 
         ports.push(Port{
@@ -398,5 +442,5 @@ unsafe fn create_ports(world: *mut LilvWorld, plugin: *const LilvPluginImpl) ->
     lilv_node_free(lv2_output_port);
     lilv_node_free(lv2_input_port);
 
-    (ports, names, n_audio_in, n_audio_out, n_atom_in)
+    Ok((ports, names, n_audio_in, n_audio_out, n_atom_in))
 }
